/* dropc.cpp
 * 
 * Copyright (C) 2010 Sylwester Arabas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <exception>
#include <iostream>
#include <vector>
using namespace std;

#define msgprefix "-- calc: "
#include "drops.hpp"

#include "ModelPseudoAdiabaticBulk.hpp"
#include "ModelPseudoAdiabaticMovingSectional.hpp"

#include "SolverCVODESAdamsFunctional.hpp"
#include "SolverCVODESBDFNewton.hpp"

#include "OutputGnuplotStuve.hpp"
#include "OutputGnuplotOneLine.hpp"

#include "UpdraftSteady.hpp"
#include "UpdraftBells.hpp"

#include "InitpTqCmdLine.hpp"
#include "InitpTqCmdLinepTRH.hpp"

#include "InBinSpectrumConstNr.hpp"

#include "InitSpectrumWhitby1978.hpp"
#include "InitSpectrumJaenicke1993.hpp"
#include "InitSpectrumODowd1997.hpp"

#include "InitBinLayoutLinear.hpp"
#include "InitBinLayoutLogLinear.hpp"

#include "SpectraMemLayoutVariable.hpp"

#include "SatVapPresMltplrKelvin.hpp"
#include "SatVapPresMltplrUnity.hpp"
#include "SatVapPresMltplrRaoult.hpp"
#include "SatVapPresMltplrRaoultKelvin.hpp"
#include "SatVapPresMltplrRaoultPetters_2007.hpp"

#include "DiffCoeffMltpltFuchsSutugin.hpp"
#include "DiffCoeffMltpltUnity.hpp"

#include "TolerancesUniform.hpp"

#include "Solutes.hpp"

#include <string>

#undef msgprefix
#define msgprefix "-- init: "

template<typename T> void str2num(char* str, T* num)
{
  stringstream ss;
  ss.exceptions(stringstream::failbit | stringstream::badbit);
  ss << str;
  try 
  {
    ss >> *num;
  }
  catch (...)
  {
    cerr << msgprefix << "failed to parse argument into a number: " << str << endl;
    throw exception();
  }
}

realtype str2real(char* str) 
{
  realtype val;
  str2num(str, &val); 
  return val; 
}

int str2int(char* str) 
{ 
  int val;
  str2num(str, &val); 
  return val; 
}

int str2bool(char* str)
{
  if (string(str) == "t") return true;
  if (string(str) == "f") return false;
  cerr << msgprefix << "failed to parse argument into a boolean value: " << str << endl;
  throw exception();
}

int main(int argc, char *argv[]) 
{

  try 
  {
    
    cerr << msgprefix << "parsing command-line options..." << endl;

    if (argc == 1) 
    {
      cerr << msgprefix << "to get help, try: " << argv[0] << " --help " << endl;
      throw exception();
    } 

    size_t model_name = 0, solver_name = 0, output_name = 0, updraft_name = 0, 
      initptq_name = 0, memlayout_name = 0, initspec_names[2] = {0, 0},
      svpmltp_name = 0, dcmltplr_name = 0, initbinl_name = 0, tol_name = 0, stop_name = 0;
    vector<bool> parsed(argc);
 
    // first walk through the options to find solver/model/output/... choices
#define arg string(argv[a])
#define handle(opt_long, opt_short, name) \
if (arg == opt_short || arg == opt_long)  { \
  if (name != 0) \
  { \
    cerr << msgprefix << opt_long << "|" << opt_short <<" specified more than once!" << endl; \
    throw exception(); \
  } \
  parsed[a] = true; \
  if (++a == argc) \
  { \
    cerr << msgprefix << opt_long << "|" << opt_short << " needs an argument" << endl; \
    throw exception(); \
  } \
  parsed[a] = true; \
  name = a; \
}
    for (size_t a = 1; a < argc; ++a)
    {
      if (arg == "--help" || arg == "-h")
      {
#undef msgprefix
#define msgprefix "-- help: "
        cerr << msgprefix << endl;
        cerr << msgprefix << "drops -- an air-parcel CCN activation model using adaptive MOL" << endl;
        cerr << msgprefix << "         based on the SUNDIALS ODE solvers" << endl;
        cerr << msgprefix << "         by Sylwester Arabas <slayoo@igf.fuw.edu.pl>" << endl;
        cerr << msgprefix << endl;
        cerr << msgprefix << "usage: " << argv[0] << " --model|-m X --solver|-s Y --output|-o Z ..." << endl;
        cerr << msgprefix << endl;
        cerr << msgprefix << "model:" << endl;
        cerr << msgprefix << "- PseudoAdiabaticMovingSectional" << endl;
        cerr << msgprefix << "dcmltplr:" << endl;
        cerr << msgprefix << "- FuchsSutugin*" << endl;
        cerr << msgprefix << "- Unity" << endl;
        cerr << msgprefix << "solver:" << endl;
        cerr << msgprefix << "- CVODESAdamsFunctional" << endl;
        cerr << msgprefix << "- CVODESBDFNewton* CVDIAG | CVDENSE* | CVDENSE-JAC | CVSPGMR | CVSPBCG | CVSPTFQMR" << endl;
        cerr << msgprefix << "stop:" << endl;
        cerr << msgprefix << "- t_max* t[s]*250" << endl;
        cerr << msgprefix << "- alt_max h[m]" << endl;
        cerr << msgprefix << "tolerances:" << endl;
        cerr << msgprefix << "- Uniform* relative_tolerance[1]*1e-8 noise_level[1]*0" << endl;
        cerr << msgprefix << "output:" << endl;
        cerr << msgprefix << "- GnuplotStuve* cloud_min[um]*1 cloud_max[um]*25 N_min[ug-1]*50 n_spec[1]*6 div_spec[1]*4 dot_step[1]*50 rec_step[1]*1 range_r_min[um]*1 range_r_max[um]*6 range_delta_t_min[K]*-.05 range_delta_t_max[K]*.125" << endl;
        //cerr << msgprefix << "- GnuplotOneLine cloud_min[um] cloud_max[um] N_min[ug-1]" << endl;
        cerr << msgprefix << "updraft:" << endl;
        cerr << msgprefix << "- Steady* U[m/s]*.25" << endl;
        cerr << msgprefix << "- SteadyPredefined .1|.25|.5|1|4|5|10" << endl;
        cerr << msgprefix << "- Bells U_max[m/s] period[s]" << endl;
        cerr << msgprefix << "initptq:" << endl;
        cerr << msgprefix << "- CmdLine p[Pa] T[K] q_v[kg/kg]" << endl;
        cerr << msgprefix << "- CmdLinepTRH* p[Pa]*100000 T[K]*280 RH[1]*.99" << endl;
        cerr << msgprefix << "memlayout:"<< endl;
        cerr << msgprefix << "- Variable* init_bins*40 max_bins*100" << endl;
        cerr << msgprefix << "- Constant n_bins" << endl;
        {
          // TODO: would be better to acquire this from the Whitby..., Jaenicke... and Solute... classes!
          string whitbyopts = "marine | clean_continental | average_background | urban";
          string jaenickeopts = "polar | background | maritime | remote_continental | desert_dust_storm | rural | urban";         
          string solutes = "NH42SO4 | NaCl | C9H14O4 | (CH2COOH)2";

          cerr << msgprefix << "initspec_0:" << endl;
          cerr << msgprefix << "- None" << endl;
          cerr << msgprefix << "- Whitby1978 " << whitbyopts << " " << solutes << endl;
          cerr << msgprefix << "- Jaenicke1993 " << jaenickeopts << " " << solutes << endl;
          cerr << msgprefix << "- ODowd1997* nss_sulphate | sea_salt* NH42SO4 | NaCl*" << endl;
          cerr << msgprefix << "- Lognormal r[um] sgma[1] N[cm-3]" << " " << solutes << endl;
          cerr << msgprefix << "initspec_1:" << endl;
          cerr << msgprefix << "- None" << endl;
          cerr << msgprefix << "- Whitby1978 " << whitbyopts << " " << solutes << endl;
          cerr << msgprefix << "- Jaenicke1993 " << jaenickeopts << " " << solutes << endl;
          cerr << msgprefix << "- ODowd1997* nss_sulphate* | sea_salt NH42SO4* | NaCl" << endl;
          cerr << msgprefix << "- Lognormal r[um] sgma[1] N[cm-3]" << " " << solutes << endl;
        }
        cerr << msgprefix << "svpmltp:" << endl;
        cerr << msgprefix << "- Raoult" << endl;
        cerr << msgprefix << "- RaoultKelvin* brent* | falsepos | bisection relative_tolerance[1]*1e-12 noise_level[1]*1e-24 max_iter[1]*20 Konopka_1996 | Petters_2007*" << endl;
        cerr << msgprefix << "initbinl:" << endl;
        cerr << msgprefix << "- Linear min[m] max[m]" << endl;
        cerr << msgprefix << "- LogLinear* min[m]*1e-9 max[m]*1e-4 max_mult[1]*2 noise_spec_conc[ug-1]*5" << endl;
        cerr << msgprefix << endl;
        throw exception();
      }
      handle("--model", "-m", model_name)
      handle("--dcmltplr", "-d", dcmltplr_name)
      handle("--solver", "-s", solver_name)
      handle("--stop", "-p", stop_name)
      handle("--tolerances", "-t", tol_name)
      handle("--output", "-o", output_name)
      handle("--updraft", "-u", updraft_name)
      handle("--initptq", "-i", initptq_name)
      handle("--memlayout", "-l", memlayout_name)
      handle("--initspec_0", "-0", initspec_names[0])
      handle("--initspec_1", "-1", initspec_names[1])
      handle("--svpmltp", "-v", svpmltp_name)
      handle("--initbinl", "-b", initbinl_name)
    }
#undef arg
#undef handle

    if (
      model_name == 0 || 
      dcmltplr_name == 0 ||
      solver_name == 0 || 
      stop_name == 0 || 
      tol_name == 0 || 
      output_name == 0 ||
      updraft_name == 0 ||
      initptq_name == 0 ||
      memlayout_name == 0 ||
      initspec_names[0] == 0 ||
      initspec_names[1] == 0 ||
      svpmltp_name == 0 ||
      initbinl_name == 0
    )
    {
      cerr << msgprefix << "a set of model/dcmltplr/solver/stop/tolerances/output/updraft/initptq/memlayout/initspec0/initspec1/svpmltp/initbinl must be specified" << endl;
      throw exception();
    }

    // instantiate the chosen solver/model/output objects
#define assureargs(offset, total, desc) \
  for (size_t a = offset + 1; a <= offset + total; ++a) \
  { \
    if (a >= argc || parsed[a]) \
    { \
      cerr << msgprefix << argv[offset - 1] << " " << argv[offset] << " requires " << total << " argument(s) (" << desc << ")" << endl; \
      throw exception(); \
    } \
    parsed[a] = true; \
  }
#define arg(idx) string(argv[idx])
#define unknown(what) \
  { \
    cerr << msgprefix << "unknown " << #what << " type: " << arg(what##_name) << endl; \
    throw exception(); \
  }

    // -------- -------- -------- -------- -------- -------- //
    // -------- -------- -------- -------- -------- -------- //

    Model *model;
    if (arg(model_name) == "PseudoAdiabaticMovingSectional")
      model = new ModelPseudoAdiabaticMovingSectional();
    else unknown(model);

    // -------- -------- -------- -------- -------- -------- //

    Updraft* updraft;
    if (arg(updraft_name) == "Steady" || arg(updraft_name) == "SteadyPredefined")
    {
      assureargs(updraft_name, 1, "U[m/s]");
      updraft = new UpdraftSteady(str2real(argv[updraft_name + 1]) * si::metres_per_second);
    }
    else if (arg(updraft_name) == "Bells")
    {
      assureargs(updraft_name, 2, "U_max[m/s] period[s]");
      updraft = new UpdraftBells(
        str2real(argv[updraft_name + 1]) * si::metres_per_second,
        str2real(argv[updraft_name + 2]) * si::seconds
      );
    }
    else unknown(updraft);

    // -------- -------- -------- -------- -------- -------- //

    quantity<si::time> t_max;
    if (arg(stop_name) == "t_max")
    {
      assureargs(stop_name, 1, "t_max[s]");
      t_max = str2real(argv[stop_name + 1]) * si::seconds;
    }
    else if (arg(stop_name) == "alt_max")
    {
      assureargs(stop_name, 1, "h_max[m]");
      t_max = updraft->t(str2real(argv[stop_name + 1]) * si::metres);
    }
    else unknown(stop);

    // -------- -------- -------- -------- -------- -------- //

    InitpTq* initptq;
    if (arg(initptq_name) == "CmdLine")
    {
      assureargs(initptq_name, 3, "p[Pa] T[K] q_v[kg/kg]");
      initptq = new InitpTqCmdLine(
        str2real(argv[initptq_name + 1]) * si::pascals,
        str2real(argv[initptq_name + 2]) * si::kelvins,
        str2real(argv[initptq_name + 3]) 
      );
    }
    else if (arg(initptq_name) == "CmdLinepTRH")
    {
      assureargs(initptq_name, 3, "p[Pa] T[K] RH[1]");
      initptq = new InitpTqCmdLinepTRH(
        str2real(argv[initptq_name + 1]) * si::pascals,
        str2real(argv[initptq_name + 2]) * si::kelvins,
        str2real(argv[initptq_name + 3]) 
      );
    }
    else unknown(initptq);

    // -------- -------- -------- -------- -------- -------- //

    size_t n_specs = 2;
    if (arg(initspec_names[0]) == "None") 
    {
      cerr << "initspec[0] must be defined" << endl;
      throw exception();
    }
    if (arg(initspec_names[1]) == "None") --n_specs;

    vector<Solute*>* svec = new vector<Solute*>(n_specs);
    vector<InitSpectrum*>* initspecs = new vector<InitSpectrum*>(n_specs);

#define handle_solute(pos) \
        if (arg(initspec_names[i] + pos) == "NH42SO4") svec->at(i) = new SoluteAmmoniumSulphate(); \
        else if (arg(initspec_names[i] + pos) == "NaCl") svec->at(i) = new SoluteSodiumChloride(); \
        else if (arg(initspec_names[i] + pos) == "C9H14O4") svec->at(i) = new SolutePinicAcid(); \
        else if (arg(initspec_names[i] + pos) == "(CH2COOH)2") svec->at(i) = new SoluteSuccinicAcid(); \
        else \
        { \
          cerr << msgprefix << "unknown Solute!" << endl; \
          throw exception(); \
        } 
    for (int i = 0; i < n_specs; ++i) 
    {
      if (arg(initspec_names[i]) == "Whitby1978") 
      {
        assureargs(initspec_names[i], 2, "name solute");
        initspecs->at(i) = new InitSpectrumWhitby1978(argv[initspec_names[i] + 1]);
        handle_solute(2)
      }
      else if (arg(initspec_names[i]) == "Jaenicke1993") 
      {
        assureargs(initspec_names[i], 2, "name solute");
        initspecs->at(i) = new InitSpectrumJaenicke1993(argv[initspec_names[i] + 1]);
        handle_solute(2)
      }
      else if (arg(initspec_names[i]) == "ODowd1997")
      {
        assureargs(initspec_names[i], 2, "name solute");
        initspecs->at(i) = new InitSpectrumODowd1997(argv[initspec_names[i] + 1]);
        handle_solute(2)
      }
      else if (arg(initspec_names[i]) == "Lognormal")
      {
        assureargs(initspec_names[i], 4, "r[um] sgma[1] N[cm-3]  solute")
        initspecs->at(i) = new InitSpectrumLognormal(
          str2real(argv[initspec_names[i] + 1]) * 1e-6 * si::metres,
          str2real(argv[initspec_names[i] + 2]),
          str2real(argv[initspec_names[i] + 3]) * pow<-3>(1e-2 * si::meters) 
        );
        handle_solute(4)
      }
      else 
      {
        cerr << msgprefix << "unknown initspec: " << argv[initspec_names[i]] << endl;
        throw exception();
      }
    }
#undef handle_solute

    // -------- -------- -------- -------- -------- -------- //

    SpectraMemLayout *memlayout;
    if (arg(memlayout_name) == "Variable")
    {
      assureargs(memlayout_name, 2, "init_bins max_bins");
      memlayout = new SpectraMemLayoutVariable(n_specs, 
        str2int(argv[memlayout_name + 1]), 
        str2int(argv[memlayout_name + 2])
      );
    }
    else if (arg(memlayout_name) == "Constant")
    {
      assureargs(memlayout_name, 1, "n_bins");
      memlayout = new SpectraMemLayoutVariable(n_specs, 
        str2int(argv[memlayout_name + 1]), 
        str2int(argv[memlayout_name + 1])
      );
    }
    else unknown(memlayout);

    // -------- -------- -------- -------- -------- -------- //

    Output *output;
    if (arg(output_name) == "GnuplotStuve")
    {
      assureargs(output_name, 11, "cloud_min[um] cloud_max[um] N_min[ug-1] n_spec[1] div_spec[1] dot_step[1] rec_step[1] range_r_min[um] range_r_max[um] range_delta_t_min[K] range_delta_t_max[K]");
      output = new OutputGnuplotStuve(model, initptq, memlayout, updraft,
        str2real(argv[output_name + 1]) * 1e-6 * si::metres,
        str2real(argv[output_name + 2]) * 1e-6 * si::metres,
        str2real(argv[output_name + 3]) / si::kilogrammes * 1e6,
        str2int(argv[output_name + 4]),
        str2real(argv[output_name + 5]),
        str2int(argv[output_name + 6]),
        str2int(argv[output_name + 7]),
        str2real(argv[output_name + 8]) * 1e-6 * si::metres,
        str2real(argv[output_name + 9]) * 1e-6 * si::metres,
        str2real(argv[output_name + 10]) * si::kelvins,
        str2real(argv[output_name + 11]) * si::kelvins
      );
    } 
    else if (arg(output_name) == "GnuplotOneLine")
    {
      assureargs(output_name, 3, "cloud_min[um] cloud_max[um] N_min[ug-1]");
      output = new OutputGnuplotOneLine(model, initptq, memlayout, updraft,
        str2real(argv[output_name + 1]) * 1e-6 * si::metres,
        str2real(argv[output_name + 2]) * 1e-6 * si::metres,
        str2real(argv[output_name + 3]) / si::kilogrammes * 1e6
      );
    }
    else unknown(output);

    // -------- -------- -------- -------- -------- -------- //

    // TODO: handle option ??
    InBinSpectrum *inbinspec;
    inbinspec = new InBinSpectrumConstNr();

    // -------- -------- -------- -------- -------- -------- //

    InitBinLayout *initbinl;
    if (arg(initbinl_name) == "Linear")
    {
      assureargs(initbinl_name, 2, "min[m] max[m]");
      initbinl = new InitBinLayoutLinear(memlayout->n_bins(0), // assumes the same number of bins in each spectrum
        str2real(argv[initbinl_name + 1]) * si::metres, 
        str2real(argv[initbinl_name + 2]) * si::metres
      );
    }
    else if (arg(initbinl_name) == "LogLinear")
    {
      assureargs(initbinl_name, 4, "min[m] max[m] max_mult[1] noise_spec_conc[ug-1]");
      initbinl = new InitBinLayoutLogLinear(memlayout->n_bins(0), // assumes the same number of bins in each spectrum
        str2real(argv[initbinl_name + 1]) * si::metres, 
        str2real(argv[initbinl_name + 2]) * si::metres,
        str2real(argv[initbinl_name + 3]),
        str2real(argv[initbinl_name + 4]) / si::kilogrammes * 1e6
      ); 
    }
    else unknown(initbinl);

    // -------- -------- -------- -------- -------- -------- //

    DiffCoeffMltplr *dcmltplr;
    if (arg(dcmltplr_name) == "Unity")
    {
      dcmltplr = new DiffCoeffMltplrUnity();
    }
    else if (arg(dcmltplr_name) == "FuchsSutugin")
    {
      dcmltplr = new DiffCoeffMltplrFuchsSutugin();
    }
    else unknown(dcmltplr);

    // -------- -------- -------- -------- -------- -------- //

    SatVapPresMltplr *svpmltp;
    if (arg(svpmltp_name) == "Unity")
    {
      svpmltp = new SatVapPresMltplrUnity();
    }
    else if (arg(svpmltp_name) == "Kelvin")
    {
      svpmltp = new SatVapPresMltplrKelvin();
    }
    else if (arg(svpmltp_name) == "Raoult")
    {
      svpmltp = new SatVapPresMltplrRaoult();
    }
    else if (arg(svpmltp_name) == "RaoultKelvin")
    {
      assureargs(svpmltp_name, 5, "name relative_tolerance[1] noise_level[1] max_iter[1] Konopka_1996|Petters_2007");
      
      SatVapPresMltplr* raoult;
      if (arg(svpmltp_name + 5) == "Konopka_1996") raoult = new SatVapPresMltplrRaoult();
      else if (arg(svpmltp_name + 5) == "Petters_2007") raoult = new SatVapPresMltplrRaoultPetters_2007();
      else 
      {
        cerr << "unknown Roult term form!" << endl;
        throw exception();
      }

      svpmltp = new SatVapPresMltplrRaoultKelvin(
        string(argv[svpmltp_name + 1]),
        str2real(argv[svpmltp_name + 2]), 
        str2real(argv[svpmltp_name + 3]),
        str2int(argv[svpmltp_name + 4]),
        raoult
      );
    }
    else unknown(svpmltp);

    // -------- -------- -------- -------- -------- -------- //

    Tolerances *tol;
    if (arg(tol_name) == "Uniform")
    {
      assureargs(tol_name, 2, "reltol[1] noise[1]")
      tol = new TolerancesUniform(
        str2real(argv[tol_name + 1]), 
        str2real(argv[tol_name + 2])
      );
    }
    else unknown(tol);

    // -------- -------- -------- -------- -------- -------- //

    ModelParams* params = model->getParams(updraft, inbinspec, memlayout, svpmltp, dcmltplr, svec);

    Solver *solver;
    if (arg(solver_name) == "CVODESBDFNewton") 
    {
      assureargs(solver_name, 1, "linear_solver")
      solver = new SolverCVODESBDFNewton(model, output, params, initptq, initspecs, 
        initbinl, memlayout, svpmltp, tol, t_max,
        string(argv[solver_name + 1])
      );
    }
    else if (arg(solver_name) == "CVODESAdamsFunctional") 
    {
      solver = new SolverCVODESAdamsFunctional(model, output, params, initptq, initspecs, 
        initbinl, memlayout, svpmltp, tol, t_max);
    }
    else unknown(solver);

    // -------- -------- -------- -------- -------- -------- //
    // -------- -------- -------- -------- -------- -------- //

#undef unknown
#undef arg
#undef assureargs

    for (size_t opt = 1; opt < parsed.size(); opt++)
    {
      if (!parsed[opt])
      {
        cerr << msgprefix << "unparsed option: " << argv[opt] << endl; 
        throw exception();
      }
    }

#undef msgprefix
#define msgprefix "-- calc: "

    // do the number crunching
    solver->run();

#undef msgprefix
#define msgprefix "-- exit: "

    // cleanup TODO: update to cover all objects... 
    //               perhaps better with auto_ptr?... 
    //               but who cares just before exit...
/*
    delete initptq;
    delete params;
    delete solver;
    delete model;
    delete output;
    delete updraft;
*/

  }
  catch (...)
  {
    cerr << msgprefix << "KO" << endl;
    exit(EXIT_FAILURE);
  }

  cerr << msgprefix << "OK" << endl;
  exit(EXIT_SUCCESS);

}
